from scrapers.linkedin_scraper import LinkedinScraper


class NoMatchingScraperType(Exception):
    pass


class ScraperFactory:
    @staticmethod
    def create(scraper_type: str, config):
        if scraper_type == "linkedin":
            return LinkedinScraper(config)
        else:
            raise NoMatchingScraperType("Incorrect scraper type")



"""Parser module will process and aggregate job posting files.
"""
import csv
import logging
import re
from configparser import NoOptionError, NoSectionError
from dataclasses import dataclass, field
from pathlib import Path

from bs4 import BeautifulSoup

from parsers.parser_base import Parser


class NoImportFiles(Exception):
    pass


class InvalidConfigArg(Exception):
    pass


class InvalidFileName(Exception):
    pass


class EmptySoup(Exception):
    "That is some bad soup"
    pass


@dataclass
class Posting:
    # config props
    input_file_name: str
    input_file: Path = Path()
    soup: BeautifulSoup = field(default=BeautifulSoup(), repr=False)

    # export props
    jobid: str = ""
    url: str = ""
    title: str = ""
    workplace_type: str = ""
    company_name: str = ""
    company_url: str = ""
    company_size: str = ""
    company_industry: str = ""
    hours: str = ""
    level: str = ""

    # used to replace asdict() as that has limited options
    def to_dict(self) -> dict:
        return {
            "jobid": self.jobid,
            "url": self.url,
            "title": self.title,
            "company_name": self.company_name,
            "workplace_type": self.workplace_type,
            "company_url": self.company_url,
            "company_size": self.company_size,
            "company_industry": self.company_industry,
            "hours": self.hours,
            "level": self.level,
        }


class LinkedinParser(Parser):
    def __init__(self, config):
        self.config = config

        # lists to hold each posting
        self.postings: list[dict] = []
        self.postings_err: list[dict] = []

        try:
            self.output_file = config.get("Parser", "output_file")
            self.output_file_errors = config.get("Parser", "output_file_err")
            self.input_dir = Path(
                Path.cwd() / config.get("Parser", "input_dir")
            )
        except (NoSectionError, NoOptionError) as e:
            logging.error(f"Err msg - {e}")
            raise e

    # parser
    def parse_export(self) -> None:
        """Export all postings to CSV file"""
        logging.info(f"Exporting to {self.output_file}")
        with open(self.output_file, "w") as f:
            writer = csv.writer(f, delimiter="|")

            writer.writerow(self.postings[0].keys())  # headers
            for posting in self.postings:
                writer.writerow(posting.values())

    # parser
    def parse_postings(self) -> None:

        if not any(self.input_dir.iterdir()):
            raise NoImportFiles("There are no files to import")

        for input_file_name in self.input_dir.iterdir():

            post = Posting(input_file_name)

            # posting - load config props
            post.input_file = self.load_posting_input_file(
                self.input_dir, input_file_name
            )
            post.soup = self.load_posting_soup(post.input_file)

            # posting - load export props
            post.jobid = self.load_posting_jobid(input_file_name)
            post.url = self.load_posting_url(post.jobid)
            post.title = self.load_posting_title(post.soup)
            post.workplace_type = self.load_posting_workplace_type(post.soup)
            post.company_name = self.load_posting_company_name(post.soup)
            post.company_url = self.load_posting_company_url(post.soup)
            (
                post.company_size,
                post.company_industry,
                post.hours,
                post.level,
            ) = self.load_posting_company_details(post.soup)

            self.postings.append(post.to_dict())

        self.parse_export()

    # input_file - config prop
    def load_posting_input_file(
        self, input_dir: Path, input_file_name: Path
    ) -> Path:
        return Path(input_dir / input_file_name)

    # soup - config prop
    def load_posting_soup(self, input_file: Path) -> BeautifulSoup:
        try:
            with open(input_file, mode="r", encoding="UTF-8") as f:
                soup = BeautifulSoup(f, "html.parser")
        except FileNotFoundError as e:
            logging.error(f"Err msg - {e}")
            raise e
        else:
            return soup

    # jobid - export prop
    def load_posting_jobid(self, input_file_name: str) -> str:
        try:
            jobid = str(input_file_name).split("_")[1]
        except IndexError as e:
            raise InvalidFileName(e) from None
        else:
            logging.info(f"{jobid} - Parsing job ")
            return jobid

    # url - export prop
    def load_posting_url(self, jobid: str) -> str:
        return "https://www.linkedin.com/jobs/view/" + jobid

    # title - export prop
    def load_posting_title(self, soup: BeautifulSoup) -> str:

        try:
            title = soup.find(class_="t-24").text.strip()

        except Exception as e:
            logging.error(f"Err msg - {e}")

            if isinstance(e, AttributeError):
                return "missing"
            return "error"
        else:
            return title

    # workplace_type - export prop
    def load_posting_workplace_type(self, soup: BeautifulSoup) -> str:

        try:
            # looking for remote (f_WT=2 in url)
            workplace_type = soup.find(
                class_="jobs-unified-top-card__workplace-type"
            ).text.strip()

        except Exception as e:
            logging.error(f"Err msg - {e}")

            if isinstance(e, AttributeError):
                return "missing"
            return "error"
        else:
            return workplace_type

    # company_name - export prop
    def load_posting_company_name(self, soup: BeautifulSoup) -> str:

        try:
            company_name = (
                soup.find("span", class_="jobs-unified-top-card__company-name")
                .find("a")
                .text.strip()
            )

        except Exception as e:
            logging.error(f"Err msg - {e}")

            # AttributeError can occur on second find()
            if isinstance(e, AttributeError):
                return "missing"
            return "error"
        else:
            return company_name

    # company_url - export prop
    def load_posting_company_url(self, soup: BeautifulSoup) -> str:

        try:
            company_url = soup.find(
                "span", class_="jobs-unified-top-card__company-name"
            ).find("a")["href"]

        except Exception as e:
            logging.error(f"Err msg - {e}")
            if isinstance(e, (AttributeError, KeyError)):
                return "missing"
            return "error"
        else:
            return company_url

    # company_details - export props
    def load_posting_company_details(self, soup: BeautifulSoup) -> tuple:

        try:
            assert type(soup) is BeautifulSoup
            company_details = soup.find_all(string=re.compile(r" · "))
        except AssertionError:
            logging.error(f"Soup should be BeautifulSoup, not {type(soup)}")
            return "error", "error", "error", "error"

        company_size: str = "missing"
        company_industry: str = "missing"
        hours: str = "missing"
        level: str = "missing"

        for section in company_details:

            section = section.strip()

            # remove leading dots if they exist
            if section[0] == "·":
                section = section[1:].strip()

            if "employees" in section and section.count(" · ") == 1:
                company_size, company_industry = section.split(" · ")
            elif "Full-time" in section and section.count(" · ") == 1:
                hours, level = section.split(" · ")

        return company_size, company_industry, hours, level



from parsers.linkedin_parser import LinkedinParser


class NoMatchingParserType(Exception):
    pass


class ParserFactory:
    @staticmethod
    def create(parser_type: str, config):
        if parser_type == "linkedin":
            return LinkedinParser(config)
        else:
            raise NoMatchingParserType("Incorrect parser type")



import json
import logging
import re
from configparser import NoOptionError, NoSectionError
from datetime import datetime
from json import JSONDecodeError
from math import ceil
from pathlib import PurePath
from time import sleep

from selenium import webdriver
from selenium.common.exceptions import (
    NoSuchElementException,
    WebDriverException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.remote.webelement import WebElement

from scrapers.scraper_base import Scraper


class InvalidCreds(Exception):
    pass


class LinkedinScraper(Scraper):
    def __init__(self, config):

        try:
            self.gecko_driver = PurePath(__file__).parent.parent / config.get(
                "Paths", "gecko_driver"
            )
            self.gecko_log = config.get("Paths", "gecko_log")
            self.creds_file = config.get("Paths", "creds")
            self.login_url = config.get("URLs", "linkedin_login")
            self.login_success = config.get("URLs", "linkedin_login_success")
            self.search_url = config.get("URLs", "linkedin_search")
            self.output_dir = config.get("Scraper", "linkedin_out_dir")
        except (NoSectionError, NoOptionError) as e:
            logging.error(f"Err msg - {e}")
            raise e

    def scrape_postings(self, postings_to_scrape: int):
        self.driver = self.browser_init(self.gecko_driver, self.gecko_log)
        username, password = self.load_creds(self.creds_file)
        self.browser_login(username, password)

        for page in range(ceil(postings_to_scrape / 25)):
            self.load_search_page(self.search_url, page * 25)
            sleep(2)  # server might reject request without wait

            logging.info("Starting to scrape")
            for i in range(25):  # 25 postings per page
                # About 7 are loaded initially.  More are loaded
                # dynamically as the user scrolls down
                postings: list = self.update_postings()
                posting: WebElement = postings[i]

                logging.info(f"Scrolling to - {i}")
                self.click_posting(posting)

                jobid = self.set_jobid(posting.get_attribute("href"))

                sleep(2)  # helps with missing content
                self.export_html(
                    self.output_dir, jobid, self.driver.page_source
                )

        logging.info("Closing browser")
        self.driver.close()

    def browser_init(self, gecko_driver, gecko_log) -> webdriver:

        logging.info("Initalizing browser")
        try:
            s = Service(executable_path=gecko_driver, log_path=gecko_log)
            driver = webdriver.Firefox(service=s)
        except WebDriverException as e:
            logging.error(f"Err msg - {e}")
            raise e

        driver.implicitly_wait(10)
        driver.set_window_size(1800, 600)

        return driver

    def load_creds(self, creds_file) -> tuple:

        logging.info("Reading in creds")
        try:
            with open(creds_file, encoding="UTF-8") as creds:
                cred_dict = json.load(creds)["linkedin"]
                username = cred_dict["username"]
                password = cred_dict["password"]
        except (FileNotFoundError, KeyError, JSONDecodeError) as e:
            logging.error(f"Err msg - {e}")
            self.driver.close()
            raise e

        return (username, password)

    def browser_login(self, username, password) -> None:

        logging.info("Navigating to login page")
        self.driver.get(self.login_url)

        logging.info(f"User name - {username}")

        logging.info("Signing in")
        try:
            self.driver.find_element(By.ID, "username").send_keys(username)
            self.driver.find_element(By.ID, "password").send_keys(password)
            self.driver.find_element(
                By.XPATH, "//button[@aria-label='Sign in']"
            ).click()

        except NoSuchElementException as e:
            logging.error(f"Err msg - {e}")
            self.driver.close()
            raise e

        if self.login_success not in self.driver.current_url:
            raise InvalidCreds

    def load_search_page(
        self, search_url, postings_scraped_total: int
    ) -> None:

        sleep(2)
        url = search_url + str(postings_scraped_total)
        logging.info(f"Loading url: {url}")
        self.driver.get(url)

    def click_posting(self, posting: WebElement) -> None:

        self.driver.execute_script(
            "arguments[0].scrollIntoView(true);",
            posting,
        )
        posting.click()

    def update_postings(self) -> list:
        # Create a list of each card (list of anchor tags).
        # Example card below:
        # <a href="/jobs/view/..." id="ember310" class="disabled ember-view
        # job-card-container__link job-card-list__title"> blah </a>
        logging.info("Updating card anchor list")
        return self.driver.find_elements_by_class_name("job-card-list__title")

    def set_jobid(self, href: str) -> str:
        # Example:
        # <a ... href="/jobs/view/2963302086/?alternateChannel...">
        try:
            jobid = re.search(r"view/(\d*)/", href)
            jobid = jobid.group(1)  # type: ignore
        except Exception as e:
            logging.error(f"Err msg - {e}")
            return "ERROR"
        else:
            return jobid  # type: ignore

    def export_html(self, output_dir, jobid: str, page_source) -> None:
        # File name syntax:
        # jobid_[JOBID]_[YYYYMMDD]_[HHMMSS].html
        # Example:
        # jobid_2886320758_20220322_120555.html
        output_file = (
            output_dir
            + "/jobid_"
            + jobid
            + "_"
            + datetime.now().strftime("%Y%m%d_%H%M%S")
            + ".html"
        )

        logging.info(f"Exporting jobid {jobid} to {output_file}")
        with open(output_file, "w+", encoding="UTF-8") as f:
            f.write(page_source)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exfill', 'exfill.parsers', 'exfill.scrapers']

package_data = \
{'': ['*'], 'exfill': ['support/*']}

install_requires = \
['bs4>=0.0.1,<0.0.2', 'selenium>=4.1.5,<5.0.0', 'setuptools-scm>=6.4.2,<7.0.0']

entry_points = \
{'console_scripts': ['script-tests = tests.runtests:main']}

setup_kwargs = {
    'name': 'exfill',
    'version': '0.1.25',
    'description': 'A small app to grab job postings from online job boards',
    'long_description': '# Introduction\n\nJob boards (like LinkedIn) can be a good source for finding job openings.  Unfortunately the search results cannot always be filtered to a usable degree.  Exfill (short for extraction) lets users scrape and parse jobs with more flexability provided by the default search.\n\nCurrently only LinkedIn is supported.\n\n# Project Structure\n\nDirectories:\n- `src/exfill/parsers` - Contains parser(s)\n- `src/exfill/scrapers` - Contains scraper(s)\n- `src/exfill/support` \n    - Contains `geckodriver` driver for FireFox which is used by Selenium\n    - Download the latest driver from the [Mozilla GeckoDriver repo in GitHub](https://github.com/mozilla/geckodriver)\n- `data/html` \n    - Not in source control\n    - Contains HTML elements for a specific job posting\n    - Populated by a scraper\n- `data/csv` \n    - Not in source control\n    - Contains parsed information in a csv table\n    - Populated by a parser\n    - Also contains an error table\n- `logs` \n    - Not in source control\n    - Contains logs created during execution\n\n## `creds.json` File\n\nSyntax should be as follows:\n\n```json\n{\n    "linkedin": {\n        "username": "jay-law@protonmail.com",\n        "password": "password1"\n    }\n}\n```\n\n# Usage\n\nThere are two `actions` required to generate usable data:\n\nFirst is the scraping action.  When called, a browser will open and perform a job query on the specified `site`.  Each posting will be exported to the `data/html` directory.  \n\nThe second action is parsing.  Each job posting in `data/html` will be opened and analyzed.  Once all postings have been analyzed a single CSV file will be exported to `data/csv`.\n\nThe csv file provides a high-level overview of all the jobs returned during the query.  When imported to a spreadsheet, users can filter on fields not present in the original search options.  Examples include sorting by companies or excluding certain industries.\n\n## Add Creds File\n\nThis is required for all usage.\n\n```bash\n# Install with git\n$ git clone git@github.com:jay-law/job-scraper.git\n\n# Create and populate creds.json.  Bash only:\ncat <<EOF > creds.json\n{\n    "linkedin": {\n        "username": "jay-law@protonmail.com",\n        "password": "password1"\n    }\n}\nEOF\n```\n\n## Use as Code\n\n```bash\n# Install with git\n$ git clone git@github.com:jay-law/job-scraper.git\n\n# Activate virtual env\n$ poetry shell\n\n# Install dependencies\n$ poetry install\n\n# Ensure creds.json exists (see above)\n\n# Execute - Scrape linkedin\n$ python3 exfill/extractor.py linkedin scrape\n\n# Execute - Parse linkedin\n$ python3 exfill/extractor.py linkedin parse\n```\n\n## Use as Module\n\n```bash\n# Install\n$ poetry add exfill\n\n# Ensure creds.json exists (see above)\n\n# Execute - Scrape linkedin\n$ python3 -m exfill.extractor linkedin scrape\n\n# Execute - Parse linkedin\n$ python3 -m exfill.extractor linkedin parse\n```\n\n# Roadmap\n\n* [x] Write unit tests\n* [ ] Improve secret handling\n* [x] Add packaging\n* [x] Move paths to config file\n* [x] Move keyword logic\n* [x] Set/include default config.ini for users installing with PIP\n* [x] Add CICD\n* [x] Automate versioning\n* [x] Add formatter (black module)\n* [x] Add static type checking (mypy module)\n* [x] Add import sorter (isort module)\n* [x] Add linter (flake8 module)\n* [x] Update string interpolation from %f to f-string\n* [x] Replace sys.exit calls with exceptions\n* [x] Update how the config object is accessed\n* [x] Migrate to `poetry` for virtual env, building, and publishing\n* [x] Replace os.path usage with pathlib\n* [x] Replace pandas export with csv export\n* [x] Replace unittest with pytest',
    'author': 'jay-law',
    'author_email': 'jay-law@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jay-law/job-scraper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
